// Universe stuff
pub const CHR_COL_NAME: &str = "chr";
pub const START_COL_NAME: &str = "start";
pub const END_COL_NAME: &str = "end";
pub const DELIMITER: char = '\t';

pub const BED_FILE_EXTENSION: &str = "bed";
pub const BAM_FILE_EXTENSION: &str = "bam";
pub const GZ_FILE_EXTENSION: &str = "gz";
pub const IGD_FILE_EXTENSION: &str = "igd";

// GTOK stuff
pub const GTOK_EXT: &str = "gtok";
